#include "serialport.h"

#include <Core/Timespec.h>
#include <fcntl.h> // ::open
#include <sys/ioctl.h> // ::ioctl
#include <termios.h> // ::tcgetattr, ::cfmakeraw, ...
#include <linux/serial.h>

#include <stdio.h>


SerialPort::SerialPort()
    : m_fd(-1)
{
}

SerialPort::~SerialPort()
{
    close();
}

bool SerialPort::open(const char * _serialPortName)
{
    int fd = -1;
    fd = ::open (_serialPortName, O_RDWR | O_NONBLOCK);
    if (fd == -1)
    {
	throw OpenSerialPortException("failed to open serial port for output");
    }
    m_fd = fd;
    return isOpen();
}

void SerialPort::close()
{
    if (isOpen())
    {
	::close(m_fd);
    }
}

bool SerialPort::isOpen(void) const
{
    return handle() != -1;
}

int SerialPort::handle() const
{
    return m_fd;
}

ssize_t SerialPort::read(char * _buffer, const size_t _size) const
{
    ssize_t r = ::read(handle(), _buffer, _size);
    if (r < 0)
    {
	perror("read");
	exit(1);
    }
    return r;
}

ssize_t SerialPort::write(const char * _buffer, const size_t _size) const
{
    return ::write(handle(), _buffer, _size);
}

static int baudrateToBaudcode(const unsigned int _baudrate)
{
    switch (_baudrate)
    {
    case       0: return B0; /* hang up */
    case      50: return B50;
    case      75: return B75;
    case     110: return B110;
    case     134: return B134;
    case     150: return B150;
    case     200: return B200;
    case     300: return B300;
    case     600: return B600;
    case    1200: return B1200;
    case    1800: return B1800;
    case    2400: return B2400;
    case    4800: return B4800;
    case    9600: return B9600;
    case   19200: return B19200;
    case   38400: return B38400;
    case   57600: return B57600;
    case  115200: return B115200;
    case  230400: return B230400;
    case  460800: return B460800;
    case  500000: return B500000;
    case  576000: return B576000;
    case  921600: return B921600;
    case 1000000: return B1000000;
    case 1152000: return B1152000;
    case 1500000: return B1500000;
    case 2000000: return B2000000;
    case 2500000: return B2500000;
    case 3000000: return B3000000;
    case 3500000: return B3500000;
    case 4000000: return B4000000;
//# define EXTA B19200
//# define EXTB B38400
    }
    return 0;
}

void SerialPort::setBaudrate(const int _baudrate)
{
    const int baudcode = baudrateToBaudcode(_baudrate);
    if (baudcode)
    {
	struct termios my_tios; /* read about all that stuff
				   with 'man cfsetospeed'    */
	::tcgetattr   (handle(), &my_tios);

	::cfsetispeed (&my_tios, baudcode);
	::cfsetospeed (&my_tios, baudcode);

	::tcsetattr (handle(), TCSANOW, &my_tios);
    }
    else
    {
	setCustomBaudrate(_baudrate);
    }
}

void SerialPort::setFormat(const int _databits, const char _parity, const int _stopbits)
{
    struct termios my_tios;
    ::tcgetattr   (handle(), &my_tios);

    my_tios.c_cflag &= ~CSIZE;
    switch (_databits)
    {
    case 5: my_tios.c_cflag |= CS5; break;
    case 6: my_tios.c_cflag |= CS6; break;
    case 7: my_tios.c_cflag |= CS7; break;
    case 8: my_tios.c_cflag |= CS8; break;
    default:
	break;
    }

    switch (_parity)
    {
    case 'N':
	my_tios.c_cflag &= ~(PARENB|PARODD);
	break;

    case 'O':
	my_tios.c_cflag |= PARENB|PARODD;
	break;

    case 'E':
	my_tios.c_cflag &= ~PARODD;
	my_tios.c_cflag |= PARENB;
	break;
    }

    if (_stopbits > 1)
    {
	my_tios.c_cflag |=CSTOPB;
    }
    else
    {
	my_tios.c_cflag &= ~CSTOPB;
    }

    ::tcsetattr (handle(), TCSANOW, &my_tios);
}

void SerialPort::setHardwareHandshake(const bool _rtscts, const bool _dsrdtr)
{
    (void)_dsrdtr;

    struct termios my_tios;
    ::tcgetattr   (handle(), &my_tios);

    if (_rtscts)
    {
	my_tios.c_cflag |= CRTSCTS;
    }
    else
    {
	my_tios.c_cflag &= ~CRTSCTS;
    }

    ::tcsetattr (handle(), TCSANOW, &my_tios);
}

void SerialPort::sendBreak(const int _duration_us)
{
    DivaCore::Timespec breakTime;
    DivaCore::Timespec remain;

    breakTime.setToNow();

    ::ioctl(handle(), TIOCSBRK, 0);

    breakTime.addNanoSeconds(1000*_duration_us);

    if (!breakTime.sleepUntil(remain))
    {
	::ioctl(handle(), TIOCCBRK, 0);
	throw SerialPortException("sleepUntil failed");
    }

    ::ioctl(m_fd, TIOCCBRK, 0);
}

void SerialPort::WaitUntilTransmitterEmpty()
{
    unsigned int result = 0;
    /* After the write returned without an error, there may
       be bytes left in the outgoing fifo or the tx-buffer
       may not be empty. We need to wait until anything is
       send out.  We may assume that the TX buffer is half
       full. We can sleep for so long and then do some
       sleeps until the TX-Buffer is realy empty.  It is not
       worth going to sleep for a shorter period.  If this
       period is to high, we may */

    while (::ioctl (handle(), TIOCSERGETLSR, &result)==0 && result != TIOCSER_TEMT)
    {
	DivaCore::Timespec waitTime;
	DivaCore::Timespec remain;

	waitTime.setToNow();
	waitTime.addNanoSeconds(1000*1000); // 1ms

	if (!waitTime.sleepUntil(remain))
	{
	    throw SerialPortException("WaitUntilTransmitterEmpty sleep failed");
	}
    }
}

void SerialPort::setCustomBaudrate(const unsigned int _newBaudrate)
{
    struct termios my_tios; /* read about all that stuff
			       with 'man cfsetospeed'    */
    struct serial_struct new_serinfo;

    if (handle()==-1) return;

    // Set up divisor for 38400 Baud
    if (::ioctl(handle(), TIOCGSERIAL, &new_serinfo) < 0)
    {
	throw OpenSerialPortException("Cannot get serial info");
    }
    
    if (new_serinfo.baud_base < _newBaudrate)
    {
	throw OpenSerialPortException("Baudbase to low can not set DMX baudrate");
    }
    
    // Set divisor to DMX Baudrate
    new_serinfo.custom_divisor = new_serinfo.baud_base / _newBaudrate;
    
    // Select custom speed for 38400 Baud.
    new_serinfo.flags &= ~ASYNC_SPD_MASK;
    new_serinfo.flags |= ASYNC_SPD_CUST;
    
    if (::ioctl(handle(), TIOCSSERIAL, &new_serinfo) < 0)
    {
	throw OpenSerialPortException("Cannot set serial info");
    }

    // Select Baudrate
    ::tcgetattr   (handle(), &my_tios);
    cfmakeraw   (&my_tios);

    /* 250KBaud 8N2, no RTSCTS */
    ::cfsetispeed (&my_tios, B38400);
    ::cfsetospeed (&my_tios, B38400);

    /* My clock has been a 22.1184MHz crystal that I have exchanged by a 24MHz crystal.
     * If it does not work, we may use a clock with half of the speed (12MHz) and use B460800.
     */
    ::tcsetattr (handle(), TCSANOW, &my_tios);
}

void  SerialPort::makeRaw()
{
    struct termios my_tios;
    ::tcgetattr   (handle(), &my_tios);
    cfmakeraw   (&my_tios);
    ::tcsetattr (handle(), TCSANOW, &my_tios);
}
