/*
 * UART Multidrop Protocol - UMP
 */

#include "uartmultidropprotocol.hpp"

#include "serialport.h"
#include "crc8_maxim.hpp"

#include <stdio.h>
#include <string.h>

#include <sys/types.h>          /* See NOTES */
#include <sys/socket.h>

#include <vector>
#include <thread>
#include <mutex>
#include <memory>
#include <functional>

#include <string>
#include <algorithm>
#include <list>

#include <stdlib.h>


namespace UartMultidropProtocol
{

    const unsigned long Frame::frameSize() const
    {
	return sizeof(Frame)-sizeof(data)+size;
    }


    class BufferReader
    {
    public:
	virtual ssize_t read(unsigned char * _buffer, const size_t _size) = 0;
    };

    class BufferWriter
    {
    public:
	class WriteErrorException // exception
	{
	public:
	    const int error;
	    WriteErrorException(const int _error) : error(_error) {}
	};

	virtual void write(const unsigned char * _buffer, const size_t _size) = 0;
    };

    class UmpTransmitter
    {
    private:
	BufferWriter & m_serialPortWriter;

	unsigned char m_frameToTransmit[20+2*256];
	size_t        m_dataSize;
	size_t        m_wrote;
	size_t        m_index;

    public:
	UmpTransmitter (BufferWriter &  _serialPortWriter)
	    : m_serialPortWriter(_serialPortWriter)
	    {
	    }

	void dumpState()
	    {
		printf ("dataSize:%lu wrote:%lu index:%lu\n", m_dataSize, m_wrote, m_index);
	    }

	void addStuffedByte(size_t & _index, unsigned char *_data, const unsigned char _src)
	    {
		if (_src == 27)
		{
		    _data[_index++] = 27;
		    _data[_index++] = 0x02;
		}
		else if (_src == 0x55)
		{
		    _data[_index++] = 27;
		    _data[_index++] = 0x05;
		}
		else
		{
		    _data[_index++] = _src;
		}
	    }

	void reset(const unsigned char *_data, const size_t _dataSize)
	    {
		if (_dataSize > 0)
		{
		    m_frameToTransmit[0] = 0x55;
		    m_frameToTransmit[1] = 0xAA;
		    m_dataSize = 2;

		    Frame &frame = *((Frame*)_data);

		    addStuffedByte(m_dataSize, m_frameToTransmit, frame.source); // source
		    addStuffedByte(m_dataSize, m_frameToTransmit, frame.dest); // destination
		    addStuffedByte(m_dataSize, m_frameToTransmit, frame.flags); // flags
		    addStuffedByte(m_dataSize, m_frameToTransmit, frame.size);

		    Crc8Maxim headerCrc8;
		    headerCrc8.update(frame.source);
		    headerCrc8.update(frame.dest);
		    headerCrc8.update(frame.flags);
		    headerCrc8.update(frame.size);
		    addStuffedByte(m_dataSize, m_frameToTransmit, headerCrc8.crc());

		    Crc8Maxim payloadCrc8;
		    for (int i=0; i<frame.size; ++i)
		    {
			payloadCrc8.update(frame.data[i]);
			addStuffedByte(m_dataSize, m_frameToTransmit, frame.data[i]);
		    }
		    addStuffedByte(m_dataSize, m_frameToTransmit, payloadCrc8.crc());
		}
		try
		{
		    m_serialPortWriter.write(m_frameToTransmit, 3);
		}
		catch(BufferWriter::WriteErrorException e)
		{
		    printf ("write-preamble error %s\n", strerror(e.error));
		}
		m_wrote = 3;
		m_index = 0;
	    }

	int input(const unsigned char _octet)
	    {
		const int TransmitChunkSize = 16;
#ifdef DEBUG
		fprintf (stderr, "<%d|%02X>", m_state, _octet); fflush(stdout);
#endif
		if (m_dataSize != 0)
		{
		    if (_octet != m_frameToTransmit[m_index])
		    {
			fprintf(stderr, "tx:error@%lu\n", m_index);fflush(stdout);
			return -1;
		    }
		    ++m_index;
		    if (m_index == m_dataSize)
		    {
			m_dataSize = 0;
			return 1; // frame transmitted successfully
		    }
		    else if (m_wrote == m_index)
		    {
			const size_t remain = m_dataSize - m_wrote;
			const size_t sizeToWrite = (remain > TransmitChunkSize) ? TransmitChunkSize : remain;
			try
			{
			    m_serialPortWriter.write(&m_frameToTransmit[m_wrote], sizeToWrite);
			}
			catch(BufferWriter::WriteErrorException e)
			{
			    printf ("write-payload error %s\n", strerror(e.error));
			}
			m_wrote += sizeToWrite;
		    }
		}
		return 0;
	    }
    };

    class UmpReceiver
    {
    private:
	enum States { Idle, ExpectPreamble1, ExpectSender, ExpectDestination, ExpectFlags, ExpectSize, ExpectHeaderCrc, ExpectBody, ExpectPayloadCrc };

	States         m_state;
	BufferWriter & m_receiveQueueWriter;

	size_t         m_index;
	int            m_stuffing;
	Crc8Maxim      m_crc8; // used for header and later for payload crc.

	Frame          m_rxframe;


    public:
	UmpReceiver (BufferWriter & _receiveQueueWriter)
	    : m_state(Idle),
	      m_receiveQueueWriter(_receiveQueueWriter)
	    {
		m_stuffing = 0;
	    }

	void reset()
	    {
		m_state = Idle;
	    }

	int byteUnstuff(unsigned char & _octet)
	    {
		if ((m_stuffing == 0) && _octet == 27)
		{
		    m_stuffing = 27;
		    return 1;
		}
		else if (m_stuffing == 27)
		{
		    m_stuffing = 0;
		    switch(_octet)
		    {
		    case 0x05:
			_octet = 0x55;
			break;
		    case 0x02:
			_octet = 27;
			break;
		    }
		}
		return 0;
	    }

	int input(unsigned char _octet)
	    {
		// printf (" (%02X) ", _octet); fflush(stdout);
		if (_octet == 0x55)
		{
		    m_state = ExpectPreamble1;
		    return 0;
		}

		if (byteUnstuff(_octet))
		{
		    return 0;
		}

		return decode(_octet);
	    }

	bool SenderAddressIsValid(const unsigned char _octet)
	    {
		// check parity (bit7) against calculated parity of bits 0..6.
		return true;
	    }

	int decode(const unsigned char _octet)
	    {
		//-- decoding
		switch (m_state)
		{
		case Idle:
		    break;

		case ExpectPreamble1:
		    if (_octet != 0xAA)
		    {
			fprintf(stderr, "rx:preamble1-error\n");
			return -1;
		    }
		    m_state = ExpectSender;
		    break;

		case ExpectSender:
		    if (!SenderAddressIsValid(_octet))
		    {
			fprintf(stderr, "rx:sender-error\n");
			return -1;
		    }
		    m_rxframe.source = _octet;
		    m_crc8 = Crc8Maxim();
		    m_crc8.update (_octet);
		    m_state = ExpectDestination;
		    break;

		case ExpectDestination:
		    m_rxframe.dest = _octet;
		    m_state = ExpectFlags;
		    m_crc8.update (_octet);
		    break;

		case ExpectFlags:
		    m_rxframe.flags = _octet;
		    m_state = ExpectSize;
		    m_crc8.update (_octet);
		    break;

		case ExpectSize:
		    m_rxframe.size = _octet;
		    m_state = ExpectHeaderCrc;
		    m_crc8.update (_octet);
		    break;

		case ExpectHeaderCrc:
		    if (m_crc8.crc() != _octet)
		    {
			fprintf(stderr, "rx:header-crc-error\n");
			return -1;
		    }
		    m_index = 0;
		    m_crc8 = Crc8Maxim();
		    m_state = ExpectBody;
		    break;

		case ExpectBody:
		    m_rxframe.data[m_index] = _octet;
		    m_crc8.update (_octet);
		    ++m_index;
		    if (m_index >= m_rxframe.size)
		    {
			m_state = ExpectPayloadCrc;
		    }
		    break;
		case ExpectPayloadCrc:
		    if (m_crc8.crc() != _octet)
		    {
			fprintf(stderr, "rx:payload-crc-error\n");
			return -1;
		    }
		    try
		    {
			m_receiveQueueWriter.write((unsigned char *)(&m_rxframe), m_rxframe.frameSize());
		    }
		    catch(BufferWriter::WriteErrorException e)
		    {
			printf ("write-frame error %s\n", strerror(e.error));
		    }
		    return 1;
		    break;

		}
		return 0;
	    }
// void writeFrameTo(BufferWriter & m_receiveQueueWriter); // writes last received frame via buffer-writer.
    };


    class UmpTransceiver
    {
    private:
	enum States { Idle, WaitLineIdle, ReceiveFrame, TransmitFrame };
	States        m_state;
	UmpReceiver     m_receiver;
	UmpTransmitter  m_transmitter;

    public:
	UmpTransceiver(BufferWriter & _serialPortWriter, BufferWriter & _receiveQueueWriter);

    public: // event interface
	bool expectTimeout(void) const;
	bool expectSerialInput(void) const;
	bool expectFrameToTransmit(void) const;

	void handleTimeout();
	void handleSerialInput(BufferReader & _reader);
	void handleFrameToTransmit(BufferReader & _reader);
    private:
	enum InputEvent { Timeout, SerialDataReceived, FrameToTransmit };
	void handleInput(const InputEvent _input, const unsigned char * _data = 0, const size_t _dataSize = 0);
    };

    UmpTransceiver::UmpTransceiver(BufferWriter & _serialPortWriter, BufferWriter & _receiveQueueWriter)
	: m_state(WaitLineIdle),
	  m_receiver(_receiveQueueWriter),
	  m_transmitter(_serialPortWriter)      
    {
    }

    bool UmpTransceiver::expectSerialInput(void) const
    {
	return true;
    }

    bool UmpTransceiver::expectFrameToTransmit(void) const
    {
	return m_state == Idle;
    }

    bool UmpTransceiver::expectTimeout(void) const
    {
	return true;
    }

    void UmpTransceiver::handleTimeout()
    {
	handleInput(Timeout);
    }

    void UmpTransceiver::handleSerialInput(BufferReader & _reader)
    {
	unsigned char buffer[100];
	int ret = _reader.read(buffer, sizeof(buffer));
	if (ret > 0)
	{
	    for (size_t i = 0; i < ret; ++i)
	    {
		handleInput(SerialDataReceived, &buffer[i], 1);
	    }
	}
    }

    void UmpTransceiver::handleFrameToTransmit(BufferReader & _reader)
    {
	unsigned char frame[100];
	int ret = _reader.read(frame, sizeof(frame));
	if (ret > 0)
	{
	    handleInput(FrameToTransmit, frame, ret);
	}
    }

    void UmpTransceiver::handleInput(const InputEvent _input, const unsigned char * _data, const size_t _dataSize)
    {
	switch (m_state)
	{
	case Idle:
	    switch (_input)
	    {
	    case Timeout:
		// not expected, can handle periodic line empty.
		break;

	    case SerialDataReceived:
		if (*_data == 0x55)
		{
		    m_state = ReceiveFrame;
		    m_receiver.reset();
		    m_receiver.input(*_data);
		}
		break;

	    case FrameToTransmit:
#ifdef DEBUG
		fprintf (stderr, " FrameToTransmit ");fflush(stdout);
#endif
		m_state = TransmitFrame;
		m_transmitter.reset(_data, _dataSize);
		break;
	    }
	    break;

	case WaitLineIdle:
	    if (_input == Timeout)
	    {
		m_state = Idle;
	    }
	    break;

	case ReceiveFrame:
	    switch (_input)
	    {
	    case Timeout:
		fprintf(stderr, "rx:timeout\n");
		m_receiver.reset();
		m_state = Idle;
		break;

	    case SerialDataReceived:
		switch (m_receiver.input(*_data))
		{
		case -1:
		    m_receiver.reset();
		    m_state = WaitLineIdle;
		    break;
		case 1:
		    // m_receiver.writeFrameTo(HandleWriter());
		    m_receiver.reset();
		    m_state = Idle;
		    break;
		case 0:
		    break;
		default:
		    break;
		}
		break;
	    default:
		break;
	    }
	    break;

	case TransmitFrame:
	    switch (_input)
	    {
	    case Timeout:
//#ifdef DEBUG
		printf (" TransmitFrame:Timeout ");
		m_transmitter.dumpState();
		fflush(stdout);
//#endif
		m_state = Idle;
		break;

	    case SerialDataReceived:
#ifdef DEBUG
		printf (" TransmitFrame:SerialDataReceived ");
		fflush(stdout);
#endif
		switch (m_transmitter.input(*_data))
		{
		case -1:
		    m_state = WaitLineIdle;
		    break;
		case 1:
		    m_state = Idle;
		    break;
		case 0:
		    break;
		default:
		    break;
		}
		break;
	    }
	    break;
	}
    }


//------ serial mainloop -----------------------------------

    class HandleReader : public BufferReader
    {
    private:
	const int m_handle;
    public:
	HandleReader(const int _handle);
	ssize_t read(unsigned char * _buffer, const size_t _size);
    };

    class HandleWriter : public BufferWriter
    {
    private:
	int m_handle;
    public:
	HandleWriter();
	void setHandle(const int _handle);
	void write(const unsigned char * _buffer, const size_t _size);
    };


    HandleReader::HandleReader(const int _handle)
	: m_handle(_handle)
    {
    }

    ssize_t HandleReader::read(unsigned char * _buffer, const size_t _size)
    {
	ssize_t ret = ::read(m_handle, _buffer, _size);
	if (ret >= 0)
	{
#ifdef DEBUG
	    for (int i=0; i < ret; ++i)
	    {
		printf ("[%d:%02X]", m_handle, _buffer[i]);
	    }
	    fflush(stdout);
#endif
	    return ret;
	}
	perror("read");
	exit(1); // TODO: throw exception instead
    }

    HandleWriter::HandleWriter()
	: m_handle(-1)
    {
    }

    void HandleWriter::setHandle(const int _handle)
    {
	m_handle = _handle;
    }

    void HandleWriter::write(const unsigned char * _buffer, const size_t _size)
    {
	size_t remain = _size;
	while (remain > 0)
	{
	    const ssize_t ret = ::write(m_handle, _buffer, _size);
	    if (ret <= 0)
	    {
		throw BufferWriter::WriteErrorException(errno);
		exit(1);
	    }
#ifdef DEBUG
	    for (int i=0; i < ret; ++i)
	    {
		printf ("{%d:%02X}", m_handle, _buffer[i]);
	    }
	    fflush(stdout);
#endif
	    remain -= ret;
	    _buffer += ret;
	}
    }

    class UmpUnixPort
    {
    private:
	SerialPort &   m_serial;
	int            m_socket;
	unsigned long  m_timeoutMicroseconds;
	HandleWriter   m_serialPortWriter;
	HandleWriter   m_receiveQueueWriter;
	UmpTransceiver m_transceiver;

    public:
	UmpUnixPort(SerialPort & _serial)
	    : m_serial(_serial),
	      m_socket(-1),
	      m_timeoutMicroseconds(5000),
	      m_transceiver(m_serialPortWriter, m_receiveQueueWriter)
	    {
	    }

	void setSocket(const int _socket)
	    {
		m_socket = _socket;
		m_receiveQueueWriter.setHandle(m_socket);
	    }

	void setTimeoutMicroseconds(const unsigned long _timeoutMicroseconds)
	    {
		m_timeoutMicroseconds = _timeoutMicroseconds;
	    }

	void run()
	    {
		m_serialPortWriter.setHandle(m_serial.handle());

		while (true)
		{
		    int maxfd = 0;
		    fd_set read_fds;

		    FD_ZERO(&read_fds);

		    if (m_transceiver.expectSerialInput())
		    {
			FD_SET(m_serial.handle(), &read_fds);
			if (m_serial.handle() > maxfd)
			    maxfd = m_serial.handle();
		    }

		    if (m_transceiver.expectFrameToTransmit())
		    {
			FD_SET(m_socket, &read_fds);
			if (m_socket > maxfd)
			    maxfd = m_socket;
		    }

		    struct timeval timeout;
		    timeout.tv_sec = 0;
		    timeout.tv_usec = m_timeoutMicroseconds;

		    const int r = select(maxfd + 1, &read_fds, 0, 0, m_transceiver.expectTimeout() ? &timeout : 0L);
		    if (r == 0)
		    {
#ifdef DEBUG
			fprintf (stderr, " t:%lu.%lu\r", timeout.tv_sec, timeout.tv_usec);
			fflush(stdout);
#endif
			m_transceiver.handleTimeout();
		    }
		    else if (FD_ISSET(m_serial.handle(), &read_fds))
		    {
			HandleReader reader(m_serial.handle());
			m_transceiver.handleSerialInput(reader);
		    }
		    else if (FD_ISSET(m_socket, &read_fds))
		    {
			HandleReader reader(m_socket);
			m_transceiver.handleFrameToTransmit(reader);
		    }
		}
	    }
    };

//-------------------------------------

    class UmpInterface_p
    {
    public:
	SerialPort  serial;
	std::string serialName;
	UmpUnixPort umpPort;
	std::thread umpPortThread;

	UmpInterface_p()
	    : umpPort(serial)
	    {
	    }
    };

    UmpInterface::UmpInterface()
	: m_p(*(new UmpInterface_p))
    {
    }

    UmpInterface::~UmpInterface()
    {
	delete & m_p;
    }

    void UmpInterface::setSerialName(const char *_serialName)
    {
	m_p.serialName = _serialName;
    }

    void UmpInterface::setFrameSocket(const int _socket)
    {
	m_p.umpPort.setSocket(_socket);
    }

    bool UmpInterface::start()
    {
	if (m_p.serialName.empty())
	{
	    printf ("serial-name\n");
	    return false;
	}

	if (!m_p.serial.open(m_p.serialName.c_str()))
	{
	    perror("open-serial");
	    return false;
	}
	m_p.serial.makeRaw();
	m_p.serial.setBaudrate(115200);
	m_p.serial.setFormat(8,'N',1);
	m_p.serial.setHardwareHandshake(false,false);

	m_p.umpPortThread = std::thread(std::bind(&UmpUnixPort::run, &m_p.umpPort));
	// m_p.umpPortThread.detach();
	m_p.umpPort.setTimeoutMicroseconds(5000);

	{
	    sched_param sch;
	    int policy;
	    pthread_getschedparam(m_p.umpPortThread.native_handle(),	&policy, &sch);
	    sch.sched_priority = 80;
	    const int e = pthread_setschedparam(m_p.umpPortThread.native_handle(), SCHED_FIFO, &sch);
	    if (e)
	    {
		printf ("Failed to setschedparam: %d : ", e);
		switch (e)
		{
		case ESRCH:  printf ("no thread"); break;
		case EINVAL: printf ("parameter"); break;
		case EPERM:  printf ("permisions"); break;
		}
		printf ("\n");
	    }
	}
	return true;
    }

}; // namespace UartMultidropProtocol
